const fs = require("fs"); 
const express = require('express');
const http = require("http");
const https = require("https");
const onFinished = require('on-finished');

const validator = require('./lib/validateConf');
const RodSession = require('./lib/rodSession');
const randomService = require('./lib/random-service');
const openAPI = require('./lib/openAPI')();
const utils = require('./lib/utils');
const maskCredential = require('./lib/maskCredential');
const q = require('./lib/queue2');

let service;
let appConf = {
    static: null,
    dynamic: null
}; 
let reqQueue = {
};
let appName="";
let app; //express
let logger;
let db;

// function _initMock(env){
//     // if(!env)throw new Error("env not found")
//     app = {}; //express
//     logger = require('./lib/tmpLogger').mock();

//     // const fs = require("fs");
//     const path = require('path');
//     const currentAppPath = path.resolve(process.cwd());

//     if( env ){
//         if (env.service ){
//             service = env.service;
//             _verifyServiceEnv(service);
//         }

//         if( env.app ){
//             appConf.dynamic = JSON.parse(env.app);
//         }
//     }

//     _loadModules(currentAppPath, app, false);

//     let rs = new RodSession( "common-rod-mock", logger, app, "common-rod-mock", _selectorServiceConf, _selectorAppConf, db);
//     let req = {
//         rodSession:{
//             detail(){
//                 return {
//                     isRawDataEnabled(){
//                         return false;
//                     }
//                 };
//             }
//         }
//     };
//     rs.commonLog( req );
//     return {
//         app: () => app,
//         this: () => rs
//     };
// }

function _init(callback, commonRodOpt={mongo_create_connect:true, detaillog_add_output_response:true, summarylog_auto_end:true}){
    app = express();

    const path = require('path');
    // const currentAppPath = mode==='mock' ?  path.resolve(process.cwd()) : path.dirname(require.main.filename);
    const currentAppPath = path.dirname(require.main.filename);
    
    //LOAD HIDE CREDENTIAL
    if ( fs.existsSync(currentAppPath + '/conf/hideCredential.json') ) { 
        console.log("FOUND \"/conf/hideCredential.json\" loading hideCredential");
        try{
            let hideCredential =  require(currentAppPath + '/conf/hideCredential.json');
            if(hideCredential.mask_template){
                Object.keys(hideCredential.mask_template).forEach(function(key) {
                    console.log("+ mask_template => " + key);
                });
            }
            if(hideCredential.mask_data){
                Object.keys(hideCredential.mask_data).forEach(function(key) {
                    if( key === 'incoming_request' || 
                        key === 'outgoing_request' || 
                        key === 'incoming_response' || 
                        key === 'outgoing_response'){
                            console.log("+ mask_data => " + key);
                    }
                });
            }
            maskCredential.init( hideCredential );
        }catch(e){
            console.error(e);
        }
        
    }

    //LOAD ENV
    if (fs.existsSync(currentAppPath + '/.env.json') && !process.env.server) { 
        console.log("FOUND \".env.json\" loading env");
        let envJSON = require(currentAppPath + '/.env.json');
        let envString = {};
        Object.keys(envJSON).forEach(function(key) {
            /** make sure it same with pm2.env */
            envString[key] = JSON.stringify(envJSON[key]);
            console.log("+ ",key);
            
        });
        Object.assign(process.env, envString);
        console.log("load env => SUCCESS");
    } 
    validator.validateConfig();
    
    if( process.env.service ){
        service = JSON.parse(process.env.service);
        _verifyServiceEnv(service);
    }

    let envCommonRod= {
        body_parser_limit:"5mb"
    };
    if( process.env.commonRod ){
        envCommonRod = Object.assign({}, envCommonRod, JSON.parse(process.env.commonRod));
    }
    
    // Middleware parser
    let bodyParser = require('body-parser');
    let collectRawData = function(req, res, buf, encoding) {
        req.rodRawData = buf.toString(encoding);
    };
    
    app.use(bodyParser.urlencoded({
        verify: collectRawData,
        limit: envCommonRod.body_parser_limit,
        extended: true
    }));
    app.use(bodyParser.json({
        verify: collectRawData,
        limit: envCommonRod.body_parser_limit,
    }));
    app.use(bodyParser.text({
        verify: collectRawData,
        type:"text/*", limit: envCommonRod.body_parser_limit
    }));
    
   
    //obtainSession + assign invoke
    app.use(function obtainSession(req, res, next) {
        if(! logger.isMetrics(req) ){
            req.invoke = req.headers["x-tid"] || randomService.generateXTid("clnt");
            if( typeof app.session === "function"){
                req.sessionID = app.session(req, res);
            }else{
                req.sessionID = "please-Init-Session";
            }
        }
        next();
    });

    // process.nextTick(()=>{
        
    //     app.use([function a(req, res, next) {
    //     },function b(req, res, next) {
    //     }]);
    //     console.log(1111, app.correlateQueue);
    //     console.log(2222,app._router.stack);
    // })

    //enqueue
    // app.use(function queueMessage(req, res, next){
    //     q.nq(req,res,next,appRodSession);
    // });

    //init before app.utils().app().const($name)
    appConf.static = require( currentAppPath + '/conf/config.json' );

    // const envServer = JSON.parse(process.env.server);
    if( process.env.commonLog ){
        let envCommonLog = JSON.parse(process.env.commonLog);
        if(envCommonLog.log)envCommonLog.log.autoAddResBody = false;
        logger = require('commonlog-kb');
        logger.init(envCommonLog, app, true, true); //true == add logger.writeApplogIncoming  and writeApplogOutgoing 
        logger.sessionID = (req,res) =>{
            return req.sessionID;
        }
        appName = envCommonLog.projectName;

    } else {
        logger = require('./lib/tmpLogger');
        logger.debug("==============================================");
        logger.debug("NOT FOUND env.commonLog, switch to console log");
        logger.debug("==============================================");
    }

    //LOAD DYNAMIC MODULES
    _loadModules(currentAppPath, app);

    if( process.env.app ){
        appConf.dynamic = JSON.parse(process.env.app);
    }
    const envServer = JSON.parse(process.env.server);
    // envServer = JSON.parse(process.env.server);
    if( envServer.health_check_url.length > 0 ){
        app.all(envServer.health_check_url,function healthCheck(req, res, next) {
            res.send("ok");
        });
    }

    // //init before app.utils().app().const($name)
    // appConf.static = require( currentAppPath + '/conf/config.json' );
    let appRodSession = new RodSession( appName, logger, app, appName, _selectorServiceConf, _selectorAppConf, null/**set db later*/);
    let rodFns = Object.getPrototypeOf(appRodSession);
    for(let rodFn in rodFns){
        // console.log('add app.' + rodFn);
        if(typeof appRodSession[rodFn] === 'function'){
            app[rodFn] = appRodSession[rodFn].bind(appRodSession);
        }
    }
    
    if( service && service.mongo && commonRodOpt.mongo_create_connect  ){
        let mongoConf = service.mongo.default;
        Promise.resolve( _connectMongodb(appRodSession, mongoConf) )
            .then(function(_db) {
                db = _db; 
                appRodSession._setDb(db);
                //global store collection, DO NOT DELETE CODE
                app._rod = {
                    collection:{
                    }
                };
                _checkFlagStart(callback, appRodSession);

            }).catch(function(err) {
                appRodSession.error(err);
            }
        );
    } else {
        _checkFlagStart(callback, appRodSession);
    }
    


    app.use(function initFn(req, res, next) {
        let rs = new RodSession( req.sessionID, logger, app, appName, _selectorServiceConf, _selectorAppConf, db);
        rs._setExpressParam(req, res, next);
        req.rodSession = rs;
        next();
    });

    if(appConf.static.append_header_resp && typeof appConf.static.append_header_resp === 'object'){
        app.use(function initAddHeader(req, res, next) {
            if(appConf.static.append_header_resp){
                res.set(appConf.static.append_header_resp);
            }
            next();
        });
    }

    function _addDataCommonLog(req, res, detaillog_add_output_response, summarylog_auto_end, closeConn, ctrlFile){
        if(detaillog_add_output_response){
            let detail = req.rodSession.detail();
            if( detail ){
                try {
                    maskCredential.outgoing_response_detail(ctrlFile.cmd, req, res, detail);
                } catch (error) {
                    console.log(error);
                }
                // v.1.3.3
                // // console.log(detail);
                // let rawData = null;
                // if( req.rodSession.detail().isRawDataEnabled() ){
                //     if(res._rodRawData){
                //         rawData = res._rodRawData;
                //     }
                //     // if(res.body){
                //     //     rawData = typeof res.body === 'string' ? res.body : JSON.stringify(res.body);
                //     // }
                // }
                // // console.log(99999, res.getHeaders());
                // let ctype = utils.checkCType( res.get('Content-Type') );
                // if( ctype !== false ){
                //     res.body = utils.convertFromType(ctype, res._rodRawData);
                //     // try {
                //     //     if(ctype === 'json'){
                //     //         res.body = JSON.parse(res._rodRawData);
                //     //     }else{
                //     //         //txt
                //     //         res.body = res._rodRawData;
                //     //     }
                //     // } catch (error) { 
                //     //     console.log(error);
                //     // }
                // }

                // // detail.addOutputResponse("client", req.rodSession.cmd, detail.InitInvoke,
                // detail.addOutputResponse("client", detail.Scenario, detail.InitInvoke,
                // rawData,
                // {
                //     StatusCode: res.statusCode,
                //     Header:res.getHeaders(),
                //     Body:res.body || null
                // });
                
                // if(typeof detail._async !== 'function' || detail._async()!==true){
                //     try{
                //         detail.end();
                //     }catch(e){
                //         req.rodSession.error(e);
                //         // console.error(1234, e);
                //     }
                // }
            } 
        }
        
        let summary = req.rodSession.summary();
        if( summary && !summary.isEnd() && summary._async!==true ){
            try {
                maskCredential.outgoing_response_summary(ctrlFile.cmd, summary);   
            } catch (error) {
                console.log(error);
            }
             
        }
         
        if( summarylog_auto_end ){
            let summary = req.rodSession.summary();
            if( summary && !summary.isEnd() && summary._async!==true ){
                if(closeConn){
                    summary.addField("SystemErrorCause","socket had closed before send the response message");
                }
                let responseResult = null;
                let responseDesc = null;
                if( res.body && typeof res.body  === 'object'){
                    responseResult = res.body.resultCode;
                    responseDesc = res.body.developerMessage;
                }
                summary.end(responseResult, responseDesc);
            }
        }

        if( logger.isDebug() ){
            //call this fn must init log with true value in the 3rd argu.
            // logger.writeApplogOutgoing(null, res);
            try {
                maskCredential.outgoing_response(ctrlFile.cmd, res, logger);
            } catch (error) {
                console.log(error);
            }
        }
    }
    
    function _addRoute(){
        if(appConf.static.routing){
            let pathObj = "conf/config.json.routing";
            Object.keys(appConf.static.routing).forEach(function(key) {
                let r = appConf.static.routing[key];
                //r == value of routing
                r.forEach((item,index) =>{
                    //item == value of aray
                    item._name = key; //naming
                    item._path = `${pathObj}.${key}[${index}]`;
    
                    ctrlFiles = _generateFileCtrl(logger, item, app, key);
                    // console.log('xxxxxxxx ',ctrlFiles);
                    
                    //Object.keys(obj)[0]
                    ctrlFiles.forEach((ctrlFile) =>{
                        if(envServer.api_docs !== false){
                            openAPI.add(ctrlFile.url, ctrlFile.method, key);
                        }

                        // let this_ctrlFile = ctrlFile.parent;
                        // this_ctrlFile.app = app;
                        // this_ctrlFile.logger = logger;
                        // this_ctrlFile.service = _selectorServiceConf;
                        // this_ctrlFile.cmd = ctrlFile.cmd;
                        // this_ctrlFile.appName = appName;
                        // this_ctrlFile.stat = logger.stat;
                       
                        let assignThis = (req, res, next) => {

                            let end = res.end;
                            res._rodIsEnd = false;
                            // It might be a little tricky here, because send supports a variety of arguments, 
                            //and you have to make sure you support all of them!
                            res.end = function (body, encoding) { 
                                try {
                                    // console.log( 999,'res.end', body,res.getHeaders(), encoding );
                                    if(Buffer.isBuffer(body)){
                                        res._rodRawData = body.toString(encoding);
                                    }else{
                                        res._rodRawData = body;
                                    }

                                    if(res._rodIsEnd){
                                        _addDataCommonLog(req, res, commonRodOpt.detaillog_add_output_response, commonRodOpt.summarylog_auto_end, true, ctrlFile)
                                        q.dq(req);
                                    }else{
                                        res._rodIsEnd = true;
                                    }
                                    end.call(this, body, encoding);
                                } catch (e) {
                                    req.rodSession.error(req.sessionID,e);
                                }
                            };
                            if(commonRodOpt.detaillog_add_output_response || 
                                commonRodOpt.summarylog_auto_end || 
                                req.rodSession._cbWaitFinished ){
                                onFinished(res, function (err, res) {
                                    try {
                                        // console.log(999,'onFinished',res._headers,res.statusCode);
                                        if(res._rodIsEnd){    
                                            _addDataCommonLog(req, res, commonRodOpt.detaillog_add_output_response, commonRodOpt.summarylog_auto_end, false, ctrlFile)
                                            q.dq(req);
                                        }else{
                                            res._rodIsEnd = true;
                                        }
                                        //CALLBACK FOR await this.waitFinished(); 
                                        if( req.rodSession._cbWaitFinished ){
                                            req.rodSession._cbWaitFinished();
                                        }
                                    } catch (e) {
                                        req.rodSession.error(req.sessionID,e);
                                    }
                                });
                            }
                            
                            req.rodSession._setCmd(ctrlFile.cmd);
                            //bind this == req
                            // ctrlFile.ctrlFile.bind(req)(req,res,next);
                            
                            if( logger.isDebug() ){
                                maskCredential.incoming_request(ctrlFile.cmd, req, logger);
                                // logger.writeApplogIncoming(req.sessionID, req);
                            }       
                            try {
                                let fc = ctrlFile.ctrlFile.call(req.rodSession, req, res, next);
                                Promise.resolve(fc).then(function(value) {
                                    if(value instanceof Error){
                                        next(value);
                                    }
                                }).catch(function(err) {
                                    next(err);
                                });
                            } catch (err) {
                                next(err)
                            }                     
                        }
                        
                        logger.info("common-rod", `URL Mapping ${ctrlFile.method} ${ctrlFile.url} => ${ctrlFile.path}`);
                        // app[ctrlFile.method](ctrlFile.url, ctrlFile.ctrlFile);

                        //enqueue
                        function queueMessage(req, res, next){
                            q.nq(req,res,next,appRodSession);
                        };
                        app[ctrlFile.method](ctrlFile.url, queueMessage, assignThis);
                    });
                    
                });
               
            });
            // openAPI.init(envServer.app_port, app);
            if(envServer.api_docs !== false){
                openAPI.init(app);
            }
            
        }
        
        handleErrorExpress( appRodSession, q );
        
    }

    function _checkFlagStart(callback, appRodSession){
        if( callback ){
            let cbBefore;
            let cbAfter;
            if( typeof callback === "function" ){
                cbBefore = callback
            }else{
                cbBefore = callback.beforeRoute;
                cbAfter = callback.afterRoute;
            }

            let beforeRouteStartFlag = cbBefore.apply(appRodSession);        
            Promise.resolve(beforeRouteStartFlag)
                .then(function(value) {
                    if(value === true){
                        _addRoute();

                        if( cbAfter ){
                            let afterRouteStartFlag = cbAfter.apply(appRodSession); 
                            Promise.resolve(afterRouteStartFlag)
                                .then(function(value) {
                                    if(value === true){
                                        _startServer(app, envServer, appRodSession);
                                    }else{
                                        appRodSession.error("callback.afterRoute return false");
                                        // process.exit(0);
                                        stopServer();
                                    }
                                }).catch(function(err) {
                                    appRodSession.error(err);
                                }
                            );
                        }else{
                            _startServer(app, envServer, appRodSession);
                        }
                        
                    } else {
                        appRodSession.error("callback.beforeRoute return false");
                        // process.exit(0);
                        stopServer();
                        
                    }
                }).catch(function(err) {
                    appRodSession.error(err);
                }
            );


        } else {
            _addRoute();
            _startServer(app, envServer, appRodSession);
        }
    }
    
    handleError();
    return app;
}

function _loadModules(currentAppPath, app, enableLogger=true ){
    let consign = require('consign');
    
    let loader = consign({
        cwd: currentAppPath,
        verbose: enableLogger,
        logger: enableLogger ? {
            log:(...x) =>{
                logger.debug("common-rod", ...x);
            },
            info: (...x) =>{
                logger.debug("common-rod", ...x);
            },
            error:(...x)=>{
                logger.error("common-rod", ...x);
            }, 
            warn:(...x)=>{
                logger.warn("common-rod", ...x)
            }
        } : undefined,
        extensions: [ '.js', '.json', '.node' ]
    }).include('src/modules')
        .then('src/services');

    if (fs.existsSync(currentAppPath+"/src/submodules")) {
        loader.then('src/submodules');
    }

    if (fs.existsSync(currentAppPath+"/src/schemas")) {
        loader.then('src/schemas');
    }
    loader.into(app);
    // RodSession.init(app);
}

function _verifyServiceEnv(service){
    Object.keys(service).forEach(function(node) {
        // let serviceName = service[key];
        Object.keys(service[node]).forEach(function(cmd) {
            // console.log(1111,cmd);
            if(cmd === "default"){
                let objToValidate = {
                    "env":{
                        "service":{
                            [node]:{
                                [cmd]:service[node][cmd]
                            }
                        }
                    }
                }
                // console.log(1111,cmd,objToValidate);
                validator.validateConType(service[node][cmd].conn_type, "env.service." + node + "." + cmd, objToValidate);
                // console.log(result);
            }
        });
    });

    //assign to complate
    Object.keys(service).forEach(function(node) {
        // let serviceName = service[key];
        Object.keys(service[node]).forEach(function(cmd) {
            if(cmd !== "default"){
                service[node][cmd] = Object.assign({}, service[node].default, service[node][cmd])
                // Object.assign(service[node][cmd], service[node].default);
            }
        });

    });
}

function _startServer(app, envServer, appRodSession){
    if(!envServer){
        envServer = JSON.parse(process.env.server);
    }
    if( envServer.maxGlobalQueue === 0 || envServer.maxSessionQueue ===0 ){
        logger.info("common-rod","disable queue message");
    }else{
        if(app.correlateQueue){
            q.init(app.correlateQueue, envServer, logger);
        }
    }

    if ((envServer.use_https === true)) {
        /* ------------- [START IMPORT SSL CONFIG] ------------ */
        const privateKey = fs.readFileSync(envServer.key);
        const certificate = fs.readFileSync(envServer.cert);
    
        const options = {
            key: privateKey,
            cert: certificate,
            ciphers: [
                "AES128-SHA256",
                "AES256-SHA256",
                "AES128-GCM-SHA256",
                "AES256-GCM-SHA384",
                "ECDHE-RSA-AES256-GCM-SHA384",
                "ECDHE-RSA-AES128-GCM-SHA256",
                "ECDHE-ECDSA-AES128-GCM-SHA256",
                "ECDHE-ECDSA-AES256-GCM-SHA384",
                "DHE-RSA-AES128-GCM-SHA256",
                "DHE-DSS-AES128-GCM-SHA256",
                "ECDHE-RSA-AES128-SHA256",
                "ECDHE-ECDSA-AES128-SHA256",
                "ECDHE-RSA-AES256-SHA384",
                "ECDHE-ECDSA-AES256-SHA384",
                "DHE-RSA-AES128-SHA256",
                "DHE-RSA-A"
            ].join(':'),
            honorCipherOrder: true,
            rejectUnauthorized: false
        };
        if(envServer.https_options){
            Object.assign( options, envServer.https_options);
        }
        
        let server = https.createServer(options, app)
        server.listen(envServer.app_port, envServer.app_host, () => {
            appRodSession.info(`Is Secure : | Port : ${envServer.app_port}`);
            require('./lib/showRoute')('', app._router.stack);
        });
        server.on('error', function (e) {
            appRodSession.error(e);
        });
    
    } else {
        let server = http.createServer(app)
        server.listen(envServer.app_port, envServer.app_host, () => {
            appRodSession.info(`Isn't Secure : | Port : ${envServer.app_port}`);
            require('./lib/showRoute')('', app._router.stack);
        });
        server.on('error', function (e) {
            appRodSession.error(e);
        });
    }
    return app;
}

async function _connectMongodb(appRodSession, mongoConf){
    let auth=""
    if(mongoConf.auth){
        auth=mongoConf.auth.user+":"+mongoConf.auth.pwd+"@";
    }

    let connectionString;
    let strLog;
    // if(mongoConf.replicaSet){
    //     let hosts = "";
    //     for (let index = 0; index < mongoConf.hosts.length; index++) {
    //         const element = mongoConf.hosts[index];
    //         if(hosts !== "")hosts += ",";

    //         hosts += element.ip;
    //         if(element.port){
    //             hosts += ":"+element.port ;
    //         }else{
    //             hosts += ":27017";
    //         }
    //     }
    //     strLog = "host=" + hosts;
    //     connectionString = `mongodb://${auth}${hosts}/${mongoConf.db}`;
    // }else{
    //     if(!mongoConf.port){
    //         mongoConf.port = "27017";
    //     }
    //     connectionString = `mongodb://${auth}${mongoConf.ip}:${mongoConf.port}/${mongoConf.db}`;
    //     strLog = `ip=${mongoConf.ip} port=${mongoConf.port}`;
    // }
    if( mongoConf.hosts !== undefined ){
        let hosts = "";
        for (let index = 0; index < mongoConf.hosts.length; index++) {
            const element = mongoConf.hosts[index];
            if(hosts !== "")hosts += ",";

            hosts += element.ip;
            if(element.port){
                hosts += ":"+element.port ;
            }else{
                hosts += ":27017";
            }
        }
        strLog = "host=" + hosts;
        connectionString = `mongodb://${auth}${hosts}/${mongoConf.db}`;
    }else{
        if(mongoConf.SRV === true){
            connectionString = `mongodb+srv://${auth}${mongoConf.ip}/${mongoConf.db}`;
            strLog = `SRV ip=${mongoConf.ip} port=${mongoConf.port}`;
        }else{
            if(!mongoConf.port){
                mongoConf.port = "27017";
            }
            connectionString = `mongodb://${auth}${mongoConf.ip}:${mongoConf.port}/${mongoConf.db}`;
            strLog = `ip=${mongoConf.ip} port=${mongoConf.port}`;
        }
        
    }

    let MongoClient = require('mongodb').MongoClient;
    let opt = mongoConf.connection_string || {};
    Object.assign( opt, {
        useNewUrlParser: true, 
        useUnifiedTopology: true,
        maxPoolSize: 10,
        serverSelectionTimeoutMS:10000 //timeout for driver connect to mongo
    })
    
    if(mongoConf.replicaSet){
        Object.assign( opt, {
            replicaSet:mongoConf.replicaSet
        })
    }
    // console.log('opt',opt);

    try {
        appRodSession.debug(`connecting to mongodb ${strLog}`);
        let client = await MongoClient.connect(connectionString, opt);
        let db = client.db();
        appRodSession.info(`connected to mongodb ${strLog} db=${mongoConf.db} opt= ${JSON.stringify(opt)} => SUCCESS`);

        //  var myobj = {
        //     name: "Company Inc",
        //     address: "Highway 37"
        // };

        // const customers1 = db.collection('customers');
        // setInterval(() => {
        //     const result = customers1.find({}, {projection:{'name': 1}}).limit(1).toArray();
        // }, 0);
        // customers1.insertOne({}, function (err, res) {
        //     if (err) throw err;
        //     console.log("1 document inserted");
        // });
        // customers1.insertOne({}, function (err, res) {
        //     if (err) throw err;
        //     console.log("1 document inserted");
        // });
        
        
    
        // const customers2 = db.collection('customers');
        // customers2.insertOne(myobj, function (err, res) {
        //     if (err) throw err;
        //     console.log("1 document inserted");
        // });
        // console.log(customers1 === customers2);
        
        // const result = await customers.find({xxxx:""}).toArray(await function (x,y){

        //     console.log(x,y);
        //     return y;
            
        // });
        // const result = await customers.find({}, {projection:{'name': 1}}).limit(1).toArray();

        // const result = await customers.aggregate([{'$facet':{'totalData':[{'$match':{}},{'$project':{'name':1,}},{'$sort':{'name':-1}},{'$limit':2},{'$skip':1}],'totalCount':[{'$match':{}},{'$group':{'_id':null,'count':{'$sum':1}}}]}}]).toArray();
        // console.log(result[0].totalData);

        // db.collection("customers").find({}, {'name': true}).toArray(function(err, results) {
        //     console.dir(results);
        // });
        
        //  var myobj = {
        //     name: "Company Inc",
        //     address: "Highway 37"
        // };
        // db.collection("customers").insertOne(myobj, function (err, res) {
        //     if (err) throw err;
        //     console.log("1 document inserted");
        //     // db.close();
        // });

        return db;
    } catch (err) {
        appRodSession.error(`connecting to mongodb ${strLog} db=${mongoConf.db} opt= ${JSON.stringify(opt)} => ERROR(${err.message})`);
        stopServer();
        // process.exit(1);
    }
}

function _selectorServiceConf(node, cmd){
    if(service[node] ){
        if(service[node][cmd]){
            return service[node][cmd];
        }else{
            return null;
        }
    }
    return null;
}

function _selectorAppConf(key, isDynamic=true){
    if( isDynamic ){
        if( appConf.dynamic ){
            return appConf.dynamic[key];
        }
    }else{
        if( appConf.static ){
            return appConf.static[key];
        }
    }
    return null;
}

function _generateFileCtrl(logger, item, app, key){
    let result = [];
    
    //force name of control 
    if( item.ctrl_name ){
        if( app.src && app.src.modules && app.src.modules[item._name] && app.src.modules[item._name][item.ctrl_name]){
            let found = false;
            let ctrlFile = app.src.modules[item._name][item.ctrl_name];
            for(let prop in ctrlFile){
                if (!ctrlFile.hasOwnProperty(prop)) continue;
                
                if('function' === typeof ctrlFile[prop]) {
                    found = {
                        ctrlFile : ctrlFile[prop],
                        parent : ctrlFile,
                        prop : prop
                    }
                    break;
                }
            }
            if(!found){
                logger.error("common-rod", `Cannot mapping ctrl file (${item._path} <--> src.modules.${key}.${item.ctrl_name}.js #not_found_fn(req, res, next))`);
                stopServer();
                // process.exit(1);
            }

            for(let i=0;i<item.method.length;i++){
                for(let j=0;j<item.url.length;j++){
                    
                    result.push({
                        ctrlFile:found.ctrlFile,
                        url:item.url[j],
                        method:item.method[i],
                        parent: found.parent,
                        // cmd : item._name,
                        //TODO test CHANGE effect
                        cmd : item.ctrl_name,
                        path:`src.modules.${key}.${item.ctrl_name}.js#${found.prop}(req, res, next)`
                    });
                }
            }

        } else {
            logger.error("common-rod", `Cannot mapping ctrl file (${item._path} <--> src.modules.${key}.${item.ctrl_name}.js)`);
            stopServer();
            // process.exit(1);
        }
        
    }else{
        //NOT FOUND CONF CTRL_NAME
        for(let i=0;i<item.method.length;i++){
            for(let j=0;j<item.url.length;j++){
                
                result.push({
                    url:item.url[j],
                    method:item.method[i]
                });
            }
        }

        for(let itr of result){
            let n = item._name;
            n = n.charAt(0).toUpperCase() + n.substring(1);
            let ctrl_name = itr.method + n;
            
            if( app.src && app.src.modules && app.src.modules[item._name] && app.src.modules[item._name][ctrl_name] ){
                let found = false;
                let ctrlFile = app.src.modules[item._name][ctrl_name];
                for(let prop in ctrlFile){
                    if (!ctrlFile.hasOwnProperty(prop)) continue;
                    
                    if('function' === typeof ctrlFile[prop]) {
                        found = {
                            ctrlFile : ctrlFile[prop],
                            parent: ctrlFile,
                            prop: prop
                        }
                        break;
                    }
                }
                if(!found){
                    logger.error("common-rod", `Cannot mapping ctrl file (${item._path} <--> src.modules.${key}.${ctrl_name}.js #not_found_fn(req, res, next))`);
                    stopServer();
                    //process.exit(1);
                }

                //update obj
                //init with 2 attr
                itr.ctrlFile = found.ctrlFile;
                itr.path = `src.modules.${key}.${ctrl_name}.js#${found.prop}(req, res, next)`;
                itr.parent = found.parent;
                // itr.cmd = item._name;
                //TODO test CHANGE effect
                itr.cmd = ctrl_name;
                //end update
    
            } else {
                logger.error("common-rod", `Cannot mapping ctrl file (${item._path} <--> src.modules.${key}.${ctrl_name}.js)`);
                stopServer();
                // process.exit(1);
            }
        }
    }
    
    
    
    return result;
}

function handleError(){
    process.on('unhandledRejection', (reason, promise) => {
        logger.warn("common-rod", 'Unhandled Rejection at', reason.stack || reason);
    });
}
function handleErrorExpress(appRodSession, q) {
    //ROUTE AT MATCH ( URL NOT FOUND DEFAULT)
    app.use((req,res,next)=>{
        q.dq(req);
        appRodSession.info( `URL NOT FOUND :: method=${req.method} url=${req.url} `);
        return res.status(403).send("URL NOT FOUND");
    });

    //EXCEED LIMIT PARSER
    app.use(function (err, req, res, next) {
        if(err.name == 'PayloadTooLargeError'){
            appRodSession.info( `${req.url} ${err.message}`);
            res.status(err.status || 413);
            res.send(err.message || 'request entity too large');
        }else{
            next(err);
        }
    });

    //CATCH ERROR
    app.use(function (err, req, res, next) {
        if (req.rodSession) {
            req.rodSession.error(req.sessionID, err);
        } else {
            console.error(err);
        }
        res.status(err.status || 500);
        res.send(err.message || 'Server Error');
    });
}

function stopServer(){
    logger.close(()=>{
        process.exit(0);
    });
}


module.exports = function(){
    return _init(...arguments);
};
// module.exports.loadModules = _loadModules;


module.exports.mock = () =>{
    return require('./lib/mock').init();
}
module.exports.mockExpressRes = () =>{
    return require('./lib/mockExpressRes')();
}
// module.exports.mock = function(){
//     // return require('./lib/mock')(...arguments);
//    return _initMock(...arguments);
// };


// module.exports.init = function(){
//     return _init(true);
// }

// module.exports.startServer = function(){
//     return _startServer(app, null, logger);
// };

module.exports.logger = ()=> logger;
