const axios = require('axios');
const buildURL = require('axios/lib/helpers/buildURL')
const randomService = require('./random-service');
const utils = require('./utils');
const maskCredential = require('./maskCredential');
const log = require('commonlog-kb');

const ERROR_TBL = {
	TIMEOUT : 'TIMEOUT',
	CONNECTION_ERROR : 'CONNECTION_ERROR'
}

// // Add a request interceptor
// axios.interceptors.request.use(function (config) {
// 	// Do something before request is sent
// 	console.log(1150,config);
// 	// config.method = "";
//     return config;
//   }, function (error) {
// 	// Do something with request error
// 	console.log(1151,error);
	
//     return Promise.reject(error);
//   });

//reuse utils of axis
const getURL = (config) => {
	return buildURL(config.url, config.params, config.paramsSerializer)
}

function InstanceHTTPReq(reqCount){
	// this.requestTime 	= Date.now();
	// this.prevOutput		= [];
	this.totalResCount 	= 0;
	this.reqCount 		= reqCount;
}

InstanceHTTPReq.prototype.recvRes 		=  function (){this.totalResCount ++;}
InstanceHTTPReq.prototype.isCompleted 	=  function (){
	return this.totalResCount >= this.reqCount;
}


//bypass parser
function transformResponse(data) {
	return data;
}

// build fn, bind this
let defaultTransformResponse = [];
for (var i = 0; i < axios.defaults.transformResponse.length; i++) {
	defaultTransformResponse.push(axios.defaults.transformResponse[i].bind(axios.defaults));
}

axios.interceptors.response.use(function (response) {
	response.rawData = response.data;
	for (var i = 0; i < defaultTransformResponse.length; i++) {
		response.data = defaultTransformResponse[i](response.data);
	}
	return response;
});

function genOptionAttribute(rodSession, optionAttribute) {
	// let {option={}} = optionAttribute;
	
	let conf = rodSession.utils().services(optionAttribute._service).conf(optionAttribute._command);
	if(!conf) throw new Error("not found config service="+optionAttribute._service+" command="+optionAttribute._command);
	optionAttribute.transformResponse = [transformResponse]; //bypass json parser
	optionAttribute.timeout = optionAttribute.timeout || (conf.timeout * 1000);
	// optionAttribute.timeout *= 1000;
	if( !optionAttribute.url ){
		let port = conf.port ? ":"+conf.port  : "" ;
		let path;
		if( optionAttribute.subUrlParams ){
			path = replaceUrlParam(rodSession, conf.path, optionAttribute.subUrlParams);
		}else{
			path = conf.path;
		}
		optionAttribute.url = `${conf.conn_type}://${conf.ip}${port}${path}`;
	}
	// if(!optionAttribute.url.startsWith(conf.conn_type)){
	// 	throw new Error("invalid[url not match with conn_type] config service="+optionAttribute._service+" command="+optionAttribute._command);
	// }

	optionAttribute.validateStatus =  false;
	optionAttribute.headers = optionAttribute.headers || {};
	if(!optionAttribute.headers['X-Tid']) {
		optionAttribute._invoke = randomService.generateXTid( rodSession.appName );
		optionAttribute.headers['X-Tid'] = optionAttribute._invoke;
	}else{
		optionAttribute._invoke = optionAttribute.headers['X-Tid'];
	}
	// optionAttribute._logHeaders = JSON.stringify(optionAttribute.headers); //to write log //10/03/2022 stringify when use
	
	// optionAttribute.option = option;
	// optionAttribute.body = optionAttribute.body || optionAttribute.option.body //case method delete want to send body

	optionAttribute._reqTime =  new Date();
	optionAttribute._maxRetry = conf.max_retry;
	// if( rodSession.detail().isRawDataEnabled() ){
		optionAttribute._rawData = null;
		if( optionAttribute.data ){
			optionAttribute._rawData = typeof optionAttribute.data === 'string' ? optionAttribute.data : JSON.stringify(optionAttribute.data);
		}
		
	// }else{
	// 	optionAttribute._rawData = null;
	// }
	return optionAttribute;
}



function genStat(rodSession, optionAttribute, isReq, isTimeout) {
	let statName = rodSession.appName + ' ';
	if (isReq) {
		statName += 'sent ' + optionAttribute._service + ' ' + optionAttribute._command + ' request';
	} else {
		statName += 'recv ' + optionAttribute._service + ' ' + optionAttribute._command + ' ';
		if (isTimeout) {
			statName += 'timeout';
		} else {
			statName += 'connection error';
		}
	}
	return statName;
}


async function requestHttp(rodSession, optionAttributes, ins) {
	if(!rodSession.detail() || !rodSession.summary()) throw new Error("not found detail OR summary");
	let requests = [];
	let requestAttributes = [];
	if(Array.isArray(optionAttributes))
		requestAttributes.push(...optionAttributes);
	else
		requestAttributes.push(optionAttributes);

	ins = ins || new InstanceHTTPReq(requestAttributes.length);
	
	function processOptionAttr(optionAttribute) {
		genOptionAttribute(rodSession, optionAttribute);

		
		let processLog = maskCredential.outgoing_request_detail(optionAttribute._service+'.'+optionAttribute._command, {
			"Header" : {...optionAttribute.headers},
			"Url" : getURL(optionAttribute),
			"QueryString" : optionAttribute.params || null,
			"Body" : optionAttribute.data, //NOT SHALLOW CLONE
			"_RawData": optionAttribute._rawData	//REMOVE AFTER CALL outgoing_request_detail
		}, rodSession.detail().isRawDataEnabled());
		let rawData = processLog._RawData;
		delete processLog._RawData;

		// console.log(9988, processLog, rawData);
		if (optionAttribute._countRetry) {
			rodSession.debug(`RETRY ${optionAttribute._countRetry}/${optionAttribute._maxRetry} to ${optionAttribute.url}`);
			rodSession.detail().addOutputRequestRetry(optionAttribute._service, optionAttribute._command, 
				optionAttribute._invoke, rawData, 
				processLog, 
				optionAttribute._countRetry, optionAttribute._maxRetry);
		} else {
			rodSession.detail().addOutputRequest(optionAttribute._service, optionAttribute._command, 
				optionAttribute._invoke, rawData, 
				processLog, 
				"http", optionAttribute.method);
			optionAttribute._countRetry = 0;
		}

		rodSession.stat(genStat(rodSession, optionAttribute, true));
		requests.push(
			axios(optionAttribute)
				.then((response)=>{
					return handleSuccess(rodSession, response, ins, optionAttribute);
				})
				.catch(async (err)=>{
					return await handleError(rodSession, err, ins, optionAttribute);
				})
		);
	};

	requestAttributes.forEach(processOptionAttr);

	//For  first loop	
	flushDetailLog(ins, rodSession);
	if(requests.length === 1){
		return await requests.pop();
	}else{
		return await axios.all(requests);
	}
}


function checkRetryCondition(err, retryCondition) {
	let retryConditions = retryCondition.split('|');
	return retryConditions.includes(err);
}


function printLog(rodSession, ins, optionAttribute, response, autoAddDetail=true ){
	let {status,headers,data,rawData} = response;
	
	if(autoAddDetail){
		if( Buffer.isBuffer(data) ){
			let ctype = utils.checkCType( headers['content-type'] );
			if( ctype !== false ){
				if(ctype === 'json'){
					data = JSON.parse(data);
				}else{
					//txt
					data = data.toString();
				}
			}
		}

		if( Buffer.isBuffer(rawData) ){
			let ctype = utils.checkCType( headers['content-type'] );
			if( ctype !== false ){
				rawData = rawData.toString();
			}else{
				//case Buffer and ctype is not TXT/JSON => show {type:"buffer" data:[xx,yy,zz]}
				rawData = JSON.stringify(rawData);
			}
		}

		let processLog = maskCredential.incoming_response_detail(optionAttribute._service+'.'+optionAttribute._command, {
			Header:{...headers},
			Body:data,
			_RawData: rawData
		});
		rawData = processLog._RawData;
		delete processLog._RawData;
		
		// console.log(9988, processLog, rawData);
		if(rodSession.detail().isRawDataEnabled()){
			rodSession.detail().addInputResponse(optionAttribute._service, optionAttribute._command, 
				optionAttribute._invoke, rawData, processLog);
		}else{
			rodSession.detail().addInputResponse(optionAttribute._service, optionAttribute._command, 
				optionAttribute._invoke, null, processLog);
		}
	}

	for (const input of rodSession.detail().Input) {
		let reqLog = maskCredential.outgoing_request(optionAttribute._service+'.'+optionAttribute._command, optionAttribute);
		let resLog = maskCredential.incoming_response(optionAttribute._service+'.'+optionAttribute._command, response, input, data, rawData);
		// let indexPrev = ins.prevOutput.findIndex(output => output.Invoke === input.Invoke);
		// if(indexPrev > -1 ){
		// 	let output = ins.prevOutput.splice(indexPrev,1)[0];
		// }
		let trancLog = 'CALL_SERVICE|'
		if(!reqLog){
			trancLog += '__TO='.concat(optionAttribute._service);
			trancLog += ' __METHOD='.concat(optionAttribute.method);
			trancLog += ' __COMMAND='.concat(optionAttribute._command);
			trancLog += ' __URL='.concat(optionAttribute.url);
			trancLog += ' __REQHEADERS='.concat(JSON.stringify(optionAttribute.headers));
			trancLog += ' __REQBODY='.concat(optionAttribute._rawData);
		}else{
			trancLog += reqLog;
		}

		trancLog += '|';

		if(!resLog){
			trancLog += ' __RESPSTATUS='.concat(status || '');
			trancLog += ' __RESPHEADER='.concat(JSON.stringify(headers)|| '');
			trancLog += ' __RESPBODY='.concat(rawData||'');
			trancLog += ' __RESPTIME='.concat(input.ResTime);
			trancLog += ' __ERRORMESSAGE='.concat(response.message || '');
		}else{
			trancLog += resLog;
		}
		
		rodSession.debug(trancLog);
	}
}

function flushDetailLog(ins, rodSession){
	if (!ins.isCompleted() ) { 
		// ins.prevOutput.push ( ...rodSession.detail().Output );
		try{
			rodSession.detail().end();
		}catch(e){
			rodSession.error(e)
		}
	}
}

async function handleSuccess(rodSession, response, ins, optionAttribute) {
	printLog(rodSession, ins, optionAttribute, response);
	response.optionAttribute = optionAttribute;
	
	
	ins.recvRes();
	flushDetailLog(ins, rodSession);
	return response;
}

async function handleError(rodSession, error, ins, optionAttribute) {
	
	let response;
	// console.log(1234131231, error.message);
	if (error.code === 'ECONNABORTED' ){
		rodSession.debug('receive TIMEOUT from', optionAttribute.url);
		rodSession.summary().addErrorBlock(optionAttribute._service, optionAttribute._command, 'ret=4', 'timeout');
		rodSession.detail().addInputResponseTimeout(optionAttribute._service, optionAttribute._command, optionAttribute._invoke);
		rodSession.stat(genStat(rodSession, optionAttribute, false, true));
			
		//response = 'TIMEOUT';
		response = ERROR_TBL.TIMEOUT;
	}else if(error.code === 'ENOTFOUND' || 
		error.code === 'ECONNRESET' || 
		error.code === 'ECONNREFUSED' || 
		error.code === 'EAI_AGAIN' ||
		error.code === 'ENETUNREACH'){
			
		rodSession.debug('receive CONNECTION_ERROR from', optionAttribute.url);
		rodSession.summary().addErrorBlock(optionAttribute._service, optionAttribute._command, 'ret=1', 'connection error');
		rodSession.detail().addInputResponseError(optionAttribute._service, optionAttribute._command, optionAttribute._invoke);
		rodSession.stat(genStat(rodSession, optionAttribute, false, false));

		// response = 'CONNECTION_ERROR';
		response = ERROR_TBL.CONNECTION_ERROR;
	} else {
		rodSession.warn('receive axios other error', error.message);
		rodSession.summary().addErrorBlock(optionAttribute._service, optionAttribute._command, 'null', error.message);
		rodSession.detail().addInputResponseError(optionAttribute._service, optionAttribute._command, optionAttribute._invoke);
		rodSession.stat(genStat(rodSession, optionAttribute, false, false));

		response = error.message 
	}
	printLog(rodSession, ins, optionAttribute, error, false);

	if (checkRetryCondition(response, rodSession.utils().services(optionAttribute._service).conf(optionAttribute._command).retry_condition)) {
		optionAttribute._countRetry ++;
		if (optionAttribute._countRetry <= optionAttribute._maxRetry) {
			// process.nextTick(async () => {
			// });
			response = await requestHttp(rodSession, optionAttribute, ins );   
		} 
	} 

	ins.recvRes();
	flushDetailLog(ins, rodSession);
	return response;
}

function replaceUrlParam(rodSession, url, params){
    let subURL = url.split('/');
    for (var i = 0; i<subURL.length; i++){
        if( subURL[i] !== "" && subURL[i].startsWith(":") ){
            let replaceValue = params[subURL[i].substring(1)];
            if( replaceValue ){
                subURL[i] =  replaceValue;
                continue;
            }else{
				rodSession.warn("this.utils().http().reqest, not found key:" + subURL[i].substring(1));
			}
        } 
    }
    return subURL.join("/");
}


module.exports.requestHttp = requestHttp;