const swaggerUi = require('swagger-ui-express');
const express = require('express');
const fs = require('fs');
const path = require('path');
const currentAppPath = path.dirname(require.main.filename);
// const pjson = require(currentAppPath + '/package.json');
let pjson = {};
if ( fs.existsSync(currentAppPath + '/package.json') ) {
    pjson = require(currentAppPath + '/package.json');
} 
// const pjson = JSON.parse( fs.readFileSync(currentAppPath + '/package.json') );

const yaml = require('js-yaml');
 

function replaceURL(url){
    if( url.includes(":") ){    
    
        let startIndex = 0;
        let lastFoundStart;
        while( true ){
            startIndex = url.indexOf(":", startIndex);
            if(startIndex === -1) break;
    
            if( url.includes("/", startIndex) ){
                lastFoundStart = url.indexOf("/", startIndex);
            }else if(url.includes("?", startIndex)){
                lastFoundStart = url.indexOf("?", startIndex);
            }else{
                lastFoundStart = url.length;
                
            }
            x = url.slice(0, lastFoundStart) + "}" + url.slice(lastFoundStart);
            x = replaceChar(x, "{", startIndex);
            // console.log(x);
    
            startIndex = lastFoundStart + 1; //+1 == }
    
            // console.log(url.length, startIndex);
            // if(startIndex >= startIndex ) break;
        }
    }
    
    return url;
}



function replaceChar(origString, replaceChar, index) {
    let firstPart = origString.substr(0, index);
    let lastPart = origString.substr(index + 1);
    
    let newString = firstPart + replaceChar + lastPart;
    return newString;
}

let openAPI = function () {
    // console.log(process.env);
    let spec = {
        "openapi": "3.0.0",
        "info": {
            "title": pjson.name,
            "description": (pjson.description || "description..."),
            "version": pjson.version
        },

        // "host": "localhost:3000",
        "servers":[
            {url: 'http://localhost:3000'}
        ],

        "tags": {

        },
        "paths":{

        }
    };
    var options = {
        // explorer: true,
        customCss: '.topbar-wrapper img {content:url("/api-docs/img/api-logo.png");}',
        swaggerOptions: {
            validatorUrl: null
        },
        customSiteTitle: "C o m m o n - R O D",
        customfavIcon: "/api-docs/img/icon.png"
        // customJs: '/custom.js'

    };

    let add = function (url, method, tag) {
        if(!spec.paths[url]){
            spec.paths[url] = {
                [method] : {
                    tags: [tag]
                }
            }
        }else{
            if(!spec.paths[url][method]){
                spec.paths[url][method] = {
                    tags : [tag]
                }
            }
        }
        // console.log(spec);
    }

    
    
    
    let init = function ( app ) {
        if ( fs.existsSync( currentAppPath + '/conf/spec.yaml' ) ) {
            try {
                const specFile = yaml.safeLoad(fs.readFileSync(currentAppPath + '/conf/spec.yaml', 'utf8'));
                if(specFile.components){
                    spec.components = specFile.components;
                }

                if(specFile.tags){
                    spec.tags = specFile.tags;
                }

                for(let url in spec.paths){
                    // url = replaceURL(url);
                    for(let method in spec.paths[url]){
                        // console.log(url, method, spec.paths[url][method]);
                        if( specFile.paths && specFile.paths[url]  ){
                            if( specFile.paths[url][method] ){
                                // console.log(specFile[url][method]);
                                delete specFile.paths[url][method].tags;
                                Object.assign(spec.paths[url][method], specFile.paths[url][method]);
                            }
                            
                        }else{
                            // console.log(99, url);
                        }
                    }
                }
            } catch(err) {
                console.error("\n*********************");
                console.error("Bad data " + currentAppPath + '/conf/spec.yaml');
                console.error(err.message);
                console.error("*********************");
            }
        }

        for(let url in spec.paths){
            for(let method in spec.paths[url]){
                if(!spec.paths[url][method].responses){
                    spec.paths[url][method].responses = {
                        "200": {
                            "description": "OK"
                        }
                    }
                }

                
                if(method !== 'get' && !spec.paths[url][method].requestBody){
                    spec.paths[url][method].requestBody = {
                        "description": "request body",
                        "required": true,
                        "content": {
                            "application/json" : {
                                "schema":{
                                    "type":"object"
                                }
                            }
                        }
                    }
                }
            }
        }
        app.use('/api-docs/img/', express.static(path.join(__dirname, '../res')));
        
        // console.log(JSON.stringify(spec, null, 4));
        app.use('/api-docs/', function(req, res, next){
            // console.log(req.url);
            if(req.originalUrl === '/api-docs/'){
                spec.servers = [{"url" : req.protocol + "://"+req.get('host')}];
                req.swaggerDoc = spec;
            }
            // else if(req.url.startsWith("/img/")){
            //     return res.sendFile( path.resolve( __dirname + "/../res/" + req.url.substring("/img/".length)) );
            // }
            next();
        }, swaggerUi.serve, swaggerUi.setup(null, options));

        // app.get('/x/x:x', function(req, res, next){
        //     console.log(req.params);
        //    res.end("11");
        // });
        
    }


    return {
        init, add
    };
}





module.exports = openAPI;