let validate = require('conf-validator');
// const e = require('express');


function validateIpOrSVC(obj, key, value){
    if(/^\d/.test(value)){
        //IP
        if(/^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/.test(value)){
            return true;
        }else{
            throw new Error("invalid IP format");
        }
        
    }else{
        //SVC
        if(/^([a-zA-Z0-9_\\.-]{3,})$/.test(value)){
            return true
        }else{
            throw new Error("invalid IP OR SVC format");
        }
    }
}

function resolve(path, obj) {
    return path.split('.').reduce(function(prev, curr) {
        return prev ? prev[curr] : null
    }, obj || self)
}

function ipOrHost(obj, key, value){
    if(key.endsWith('.ip')){
        if("HOSTS" === value.toUpperCase()){
            return true;
        }else{
            if(/^\d/.test(value)){
                //IP
                if(/^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/.test(value)){
                    return true;
                }else{
                    throw new Error("invalid IP format");
                }
                
            }else{
                //SVC
                if(/^([a-zA-Z0-9_\\.-]{3,})$/.test(value)){
                    return true
                }else{
                    throw new Error("invalid IP OR SVC format");
                }
            }
        }
    }else{
        //HOSTS
        // let port = resolve(key.substring(0,key.lastIndexOf("."))+".port", obj);
        // console.log(9988899);
        return true;
    }
    
}

function validateIpOrSVCWithArray(obj, key, value){
    if( Array.isArray(value) ) {
        let port = resolve(key.substring(0,key.lastIndexOf("."))+".port", obj);
        if(port.length !== value.length){
            throw new Error( "port is not equals" );
        }
        for (let index = 0; index < value.length; index++) {
            const element = value[index];
            if(/^\d/.test(element)){
                //IP
                if(! /^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/
                .test(element)) return false;
            }else{
                //SVC
                if(!/^([a-zA-Z0-9_\\.-]{3,})$/.test(element))return false;
            }
        }
        return true;
    }else{
        if(/^\d/.test(value)){
            //IP
            return /^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/
                .test(value);
        }else{
            //SVC
            return /^([a-zA-Z0-9_\\.-]{3,})$/
                .test(value);
        }
    }
}

function validateConfig(){
    const schema = [{
        "name": "validate_env_server",
        "value": validate.buildPM2ENVObj(),
        schema: [{
                "config_path": "env.server.app_host",
                "array": false,
                "required_attr": true,
                "required_val": true,
                "type": "condition",
                "condition" : validateIpOrSVC
            }, {
                "config_path": "env.server.app_port",
                "array": false,
                "required_attr": true,
                "required_val": true,
                "type": "port"
            }, {
                "config_path": "env.server.https_options",
                "array": false,
                "required_attr": false,
                "required_val": true,
                "type": "object"

            }, {
                "config_path": "env.server.use_https",
                "array": false,
                "required_attr": false,
                "required_val": true,
                "type": "condition",
                "condition": (obj, key, value)=>{
                    if(!(typeof value === "boolean")){
                        throw new Error( "this config must be type boolean" );
                    }
                    if(value === true){
                        if(obj.env.server.cert===undefined || !(typeof obj.env.server.cert === "string") ){
                            throw new Error( "env.server.use_https=true, env.server.cert is required" );
                        }
                        if(obj.env.server.key===undefined || !(typeof obj.env.server.key === "string") ){
                            throw new Error( "env.server.use_https=true, env.server.key is required" );
                        }
                    }
                    return true;
                }

            },{
                "config_path": "env.server.cert",
                "array": false,
                "required_attr": false,
                "required_val": false,
                "type": "string"
                
            }, {
                "config_path": "env.server.key",
                "array": false,
                "required_attr": false,
                "required_val": false,
                "type": "string"                
            }, {
                "config_path": "env.server.health_check_url",
                "array": true,
                "required_attr": false,
                "required_val": false,
                "type": "string"
                
            }
            
            
        ]
    }];

    let isCorrect = validate.validator(schema,(txt)=>{
        console.log(txt);
    });
    if (!isCorrect) {
        console.log('Validate Config Fail');
        process.exit(1);
    }
    
    return isCorrect;
}


function validateConType(conn_type, path, obj, name){
    // console.log(JSON.stringify(obj));
    let schema = [];
    if(conn_type === "http" || conn_type === "https" ){
        schema.push({
            "name": "validate_env_service_http" ,
            "value": obj,
            schema: [{
                    "config_path": `${path}.conn_type`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "conn_type"
                }, {
                    "config_path": `${path}.ip`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "condition",
                    "condition" : validateIpOrSVC
                }, {
                    "config_path": `${path}.port`,
                    "array": false,
                    "required_attr": true,
                    "required_val": false,
                    "type": "port"
                },  {
                    "config_path": `${path}.path`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "string"
                }, {
                    "config_path": `${path}.max_retry`,
                    "array": false,
                    "required_attr": false,
                    "required_val": false,
                    "type": "integer"
                },{
                    "config_path": `${path}.timeout`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "integer"
                },{
                    "config_path": `${path}.retry_condition`,
                    "array": false,
                    "required_attr": false,
                    "required_val": false,
                    "type": "string"
                }
            ]
        });
    }else if(conn_type === "mongodb"){
        let schemaForMongo =  {
            "name": "validate_env_service_mongodb" ,
            "value": obj,
            schema: [{
                    "config_path": `${path}.conn_type`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "conn_type"
                }, {
                    "config_path": `${path}.ip`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "condition",
                    "condition" : validateIpOrSVC
                }, {
                    "config_path": `${path}.port`,
                    "array": false,
                    "required_attr": false,
                    "required_val": false,
                    "type": "port"
                }, {
                    "config_path": `${path}.replicaSet`,
                    "array": false,
                    "required_attr": false,
                    "required_val": true,
                    "type": "string"
                }, {
                    "config_path": `${path}.db`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "string"
                }, {
                    "config_path": `${path}.max_retry`,
                    "array": false,
                    "required_attr": false,
                    "required_val": false,
                    "type": "integer"
                },{
                    "config_path": `${path}.timeout`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "integer"
                },{
                    "config_path": `${path}.retry_condition`,
                    "array": false,
                    "required_attr": false,
                    "required_val": false,
                    "type": "string"
                },{
                    "config_path": `${path}.SRV`,
                    "array": false,
                    "required_attr": false,
                    "required_val": false,
                    "type": "boolean"
                }
            ]
        };
        
        // let replicaSet = resolve(path+".replicaSet", obj);
        // if(replicaSet){
        //     schemaForMongo.schema.splice(1,2); //delete ip / port
        //     schemaForMongo.schema.splice(1, 0, {
        //         "config_path": `${path}.hosts`,
        //         "array": true,
        //         "required_attr": true,
        //         "required_val": true,
        //         "type": "object",
        //     },{
        //         "config_path": `${path}.hosts[].ip`,
        //         "array": false,
        //         "required_attr": true,
        //         "required_val": true,
        //         "type": "condition",
        //         "condition" : validateIpOrSVC
        //     },{
        //         "config_path": `${path}.hosts[].port`,
        //         "array": false,
        //         "required_attr": false,
        //         "required_val": false,
        //         "type":"port"
        //     },{
        //         "config_path": `${path}.replicaSet`,
        //         "array": false,
        //         "required_attr": true,
        //         "required_val": true,
        //         "type":"string"
        //     });
        // }
        // schema.push(schemaForMongo);
        let hosts = resolve(path+".hosts", obj);
        if(hosts !== undefined){
            schemaForMongo.schema.splice(1,2); //delete ip / port
            schemaForMongo.schema.splice(1, 0, {
                "config_path": `${path}.hosts`,
                "array": true,
                "required_attr": true,
                "required_val": true,
                "type": "object",
            },{
                "config_path": `${path}.hosts[].ip`,
                "array": false,
                "required_attr": true,
                "required_val": true,
                "type": "condition",
                "condition" : validateIpOrSVC
            },{
                "config_path": `${path}.hosts[].port`,
                "array": false,
                "required_attr": false,
                "required_val": false,
                "type":"port"
            });
        }
        schema.push(schemaForMongo);

    }else if(conn_type === "redis"){
        schema.push({
            "name": "validate_env_service_redis" ,
            "value": obj,
            schema: [{
                    "config_path": `${path}.conn_type`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "conn_type"
                }, {
                    "config_path": `${path}.host`,
                    "array": false,
                    "required_attr": true,
                    "required_val": false,
                    "type": "string"
                }, {
                    "config_path": `${path}.port`,
                    "array": false,
                    "required_attr": true,
                    "required_val": false,
                    "type": "port"
                }, {
                    "config_path": `${path}.max_retry`,
                    "array": false,
                    "required_attr": false,
                    "required_val": false,
                    "type": "integer"
                },{
                    "config_path": `${path}.timeout`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "integer"
                },{
                    "config_path": `${path}.retry_condition`,
                    "array": false,
                    "required_attr": false,
                    "required_val": false,
                    "type": "string"
                }
            ]
        });

    }else if(conn_type === "smtp"){
        schema.push({
            "name": "validate_env_service_smtp" ,
            "value": obj,
            schema: [{
                    "config_path": `${path}.conn_type`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "conn_type"
                }, {
                    "config_path": `${path}.host`,
                    "array": false,
                    "required_attr": true,
                    "required_val": false,
                    "type": "string"
                }, {
                    "config_path": `${path}.port`,
                    "array": false,
                    "required_attr": true,
                    "required_val": false,
                    "type": "port"
                }, {
                    "config_path": `${path}.secure`,
                    "array": false,
                    "required_attr": true,
                    "required_val": true,
                    "type": "boolean"
                },{
                    "config_path": `${path}.auth`,
                    "array": false,
                    "required_attr": false,
                    "required_val": true,
                    "type": "condition",
                    "condition": (obj, key, value)=>{
                        if(!(typeof value === "object")){
                            throw new Error( "this config must be type object" );
                        }
                        
                        if(value.user ===undefined || !(typeof value.user === "string") || value.user.length === 0 ){
                            throw new Error( key+".user is required" );
                        }
                        if(value.pass ===undefined || !(typeof value.pass === "string") || value.pass.length === 0 ){
                            throw new Error( key+".pass is required" );
                        }
                        return true;
                    }
                }
                
                // , {
                //     "config_path": `${path}.auth.user`,
                //     "array": false,
                //     "required_attr": false,
                //     "required_val": true,
                //     "type": "string"
                // }, {
                //     "config_path": `${path}.auth.pass`,
                //     "array": false,
                //     "required_attr": false,
                //     "required_val": true,
                //     "type": "string"
                // } 


                
            ]
        });

    } else {
        console.error('validateConType conn_type not support');
        console.error('Validate Config Fail');
        process.exit(1);
    }
    
    let isCorrect = validate.validator(schema)
    if (!isCorrect) {
        console.error('Validate Config Fail');
        process.exit(1);
    }
    
    return isCorrect;
}


exports.validateConfig = validateConfig;
exports.validateConType = validateConType;
